/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	FASTCONV.C (Hauptprogramm)				*/
/*  Version  :	(s.u.)							*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:	Schnelle Faltung von Zeitsignalen mit FFT		*/
/*									*/
/*  history  :	29.6.93	Beginn		 				*/
/*		29.6.93/V1.3						*/
/*		30.6.93/V2.0 Rumpfgeschwindigkeit gesteigert durch	*/
/*			     Verzicht auf Fifos, fft schneller		*/
/*		30.6.93/V2.1 Verdopplung der Geschwindigkeit durch	*/
/*			     Nutzung Imaginarteil in Zeitsignal f. FFT	*/
/*		1.7.93/V2.2  OUTBUFFSIZE-Mimik durch CHUNKSIZEBYTES	*/
/*			     (s.u.) bei fwrite() sichergestellt.	*/
/*		(Fortsetzung s.u.)					*/
/*									*/
/************************************************************************/
/*
 *  $Source: /home/mummert/src/SWMUMDIS/fastconv/RCS/fastconv.c,v $
 *  $Revision: 2.10 $
 *  $Date: 1999/08/04 14:00:35 $
 *  $Author: mummert $
 *  $Log: fastconv.c,v $
 *  Revision 2.10  1999/08/04 14:00:35  mummert
 *  Changed caller interface of module FFT (FFT now allows multiple use)
 *
 *  Revision 2.9  1999/07/24 00:57:23  mummert
 *  letzte Zahl bei ASCII-Lesen richtig, OFIELDFMT-Option implementiert
 *
 *  Revision 2.8  1999/02/25 22:44:25  mummert
 *  New freeware-licence; module portab with MS-Visual-C++ support
 *
 *  Revision 2.7  1998/10/20 00:28:20  mummert
 *  Nochmals verbesserte Portabilitaet, Module PORTAB und FFT
 *
 *  Revision 2.6  1998/10/14 01:05:29  mummert
 *  Weiter verbesserte Portabilitaet
 *
 *  Revision 2.5  1998/10/11 23:48:07  mummert
 *  Verbesserung Portabilitaet
 *
 *  Revision 2.4  1998/08/28 00:49:59  mummert
 *  Geaendertes Modul KEY mit DFLTPARMPATH; DFLTPARMFILE nun fastconv.par
 *
 *  Revision 2.3  1998/08/19 20:26:14  mummert
 *  RCS-Programmkopf; ld(N) -> log2(N) in info;
 *  veraenderte Module PORTAB, KEY, FFT
 *
 */
	
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "portab.h"			/* Unix/Atari Portabilitaet	*/
#include "key.h"			/* Keytab-Management-Modul	*/
#include "fft.h"

#define CHUNKSIZEBYTES	512		/* Groesse in Bytes fuer call 	*/
					/* fwrite(), bei der file-buff.	*/
					/* noch nicht geflushed wird,	*/
					/* obwohl durch setvbuff() viel	*/
					/* groesser. Haengt wohl von	*/
					/* u-area-buffergroesse ab ???	*/
int	DEBUG = 0;

#define	DFLTPARMFILE	"fastconv.par"	/* Default Eingabedatei		*/

/*----------------------------------------------------------------------*/
/*	globale Variable						*/
/*----------------------------------------------------------------------*/

char 	g_ifile[MAXLINE],		/* Eingangsdatei		*/
	g_ofile[MAXLINE],		/* Ausgangsdatei		*/
	g_cfile[MAXLINE],		/* Faltungsdatei		*/
	g_cfmt[MAXLINE],		/* Faltungsformat		*/
	g_ifmt[MAXLINE],		/* Eingangsformat		*/
	g_ofmt[MAXLINE],		/* Ausgangsformat		*/
	g_ofieldfmt[MAXLINE],		/* printf()-Format bei ascii	*/
	g_debug[MAXLINE];		/* Debug-Schalter		*/

double  g_scale,			/* Verstaerkung			*/
	g_fillup,			/* Fuellwert fuer last block	*/
	g_fa;				/* Abtastfrequenz Hz		*/

int	g_ldn,				/* ld(N) von Blockgroesse	*/
	g_overlap,			/* Overlap			*/
	g_outbuffsize;			/* Ausgangspuffergroesse	*/

char 	*g_stdoutbuf;			/* Ausgangspufferzeiger		*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
					/* Keytab (vergl. key.h/c)	*/
KEY_TAB kt[]= {
"IFILE",	g_ifile,	STRING,	"Datei Eingangssignal (stdin : '-')",
				"-",	EMPTY,
"IFMT",		g_ifmt,		STRING,	"Dateiformat",
				"word",	"ubyte byte word double float ascii",
"LDN",		pt(g_ldn),	INT,	"log2(N) von Blockgroesse N",
				"9",	"1 16",
"CFILE",	g_cfile,	STRING,	"finites Faltungssignal (stdin : '-')",
				"-",	EMPTY,
"CFMT",		g_cfmt,		STRING,	"Format (signifkante Laenge < K)",
				"word",	"ubyte byte word double float ascii",
"OVERLAP",	pt(g_overlap),	INT,	"Ueberlappung K (K < N, K=0 -> K=N/5)",
				"0",	"0 65535",
"SCALE",	pt(g_scale),	DOUBLE,	"Verstaerkungsfaktor",
				"3.051851E-5", EMPTY,	
"OFILE",	g_ofile,	STRING, "Datei Ausgangssignal (stdout : '-')",
				"-",	EMPTY,
"OFMT",		g_ofmt,		STRING,	"Dateiformat",
				"word",	"ubyte byte word double float ascii",
"OFIELDFMT",	g_ofieldfmt,	STRING,	"OFMT=ascii: printf-Format",
				"%+le",	EMPTY,
"OUTBUFFSIZE",	pt(g_outbuffsize),	INT,
				"Ausgangsdateipuffer in Bytes",
				"1024000", "0 12345678",
"FILLUP",	pt(g_fillup),	DOUBLE,	"Fuellwert Block",
				"0",	EMPTY,
"DEBUG",	g_debug,	STRING,	"Debug-Ausgabe",
				"nein",	"j n ja nein",
0L,0L,0L,0L,0L,0L };

/*----------------------------------------------------------------------*/
/*  main								*/
/*----------------------------------------------------------------------*/
main(argc, argv)
int argc;
char *argv[];
{
	char *errmsg;
	static char afmt[MAXLINE+2];
	double *in, *out;
	FILE *istream, *ostream, *cstream;

	register int i;
	FFTELEM *xr, *xi, *cfr, *cfi, zr, zi;
	FFTDATA *fftp;
	int fftsize, buffsize, found;

					/* Keytab-Mechanismus		*/
	if (initkeytab(argc, argv, kt, DFLTPARMFILE))
		exit(-1);
	if ((g_debug[0] | 0x20) == 'j')
		DEBUG = 1;

	strcpy(afmt, g_ofieldfmt);
	strcpy(afmt+strlen(g_ofieldfmt), "\n");

	fftsize = 1 << g_ldn;
	buffsize = fftsize * 2;

	if (g_overlap <= 0)
		g_overlap = fftsize/5;
	if (g_overlap >= fftsize) {
		errmsg = "Overlap K >= N";
		errorhandler(errmsg);
		exit(-1);
	}
 
	if (DEBUG) {
 	    fprintf(stderr,"fftsize N: %d\n", fftsize);
 	    fprintf(stderr,"overlap K: %d\n", g_overlap);
 	}

	if ((in  = (double*)(malloc(sizeof(double)*buffsize))) == NULL ||
	    (out = (double*)(malloc(sizeof(double)*buffsize))) == NULL ||
	    (cfi = (FFTELEM*)(malloc(sizeof(FFTELEM)*fftsize))) == NULL ||
	    (cfr = (FFTELEM*)(malloc(sizeof(FFTELEM)*fftsize))) == NULL ||
	    (g_stdoutbuf = malloc(g_outbuffsize)) == NULL ||
	    (fftp = fft_init(g_ldn)) == NULL) {
		errmsg = "nicht genug Speicher";
		errorhandler(errmsg);
		exit(-1);
	}
	xr = fftp->r;
	xi = fftp->i;
					/* ggf. Eingabedatei oeffnen	*/
	if (g_ifile[0] != '-' || g_ifile[1] != 0) {
		if ((istream = fopen(g_ifile, READ)) == NULL) {
			errmsg = catmsg(g_ifile, "kann nicht oeffnen");
			errorhandler(errmsg);
			exit(-1);
		}
	} else {
		istream = stdin;
	}
					/* Faltungsdatei oeffnen	*/
	if (g_cfile[0] != '-' || g_cfile[1] != 0) {
		if ((cstream = fopen(g_cfile, READ)) == NULL) {
			errmsg = catmsg(g_cfile, "kann nicht oeffnen");
			errorhandler(errmsg);
			exit(-1);
		}
	} else {
		cstream = stdin;
		if (cstream == istream) {
			errmsg = "IFILE == CFILE nicht bei stdin";
			errorhandler(errmsg);
			exit(-1);
		}
	}
					/* ggf. Ausgabedatei oeffnen	*/
	if (g_ofile[0] != '-' || g_ofile[1] != 0) {
		if ((ostream = fopen(g_ofile, WRITE)) == NULL) {
			errmsg = catmsg(g_ofile, "kann nicht oeffnen");
			errorhandler(errmsg);
			exit(-1);
		}
	} else {
		ostream = stdout;
	}
					/* Ausgabepuffer vergroessern	*/
	setvbuf(ostream, g_stdoutbuf, _IOFBF, g_outbuffsize);
					/* Faltungsfunktion lesen	*/
					/* FFT von Faltungsfunktion	*/
	streamread(cstream, g_cfmt, in, fftsize, g_fillup);
	for (i = 0; i < fftsize; i++) {
		xr[i] = (FFTELEM)(in[i] * g_scale);
		xi[i] = 0;
	}
	fft(fftp); 
	for (i = 0; i < fftsize; i++) {
		cfr[i] = xr[i];
		cfi[i] = xi[i];
	}
					/* Nichtsignifikante Werte vor	*/
					/* Eingangssignal fuegen	*/
	for (i = 0; i < g_overlap; i++)
		in[i] = g_fillup;
					/* Schleife bis Signaldateiende	*/
	while (!feof(istream)) {
	    found = streamread(istream, g_ifmt,
		in+g_overlap, buffsize-2*g_overlap, g_fillup);
						/* FFT-Felder Besetzen	*/
	    for(i = 0; i < fftsize; i++) {
		xr[i] = (FFTELEM)(in[i]);
		xi[i] = (FFTELEM)(in[i+fftsize-g_overlap]);
	    }
						/* schnelle Faltung	*/
	    fft(fftp);
	    for (i = 0; i < fftsize; i++) {
		zr = xr[i] * cfr[i] - xi[i] * cfi[i];
		zi = xr[i] * cfi[i] + xi[i] * cfr[i];
		xr[i] = zr;
		xi[i] = zi;
	    }
	    ifft(fftp);
					/* Umspeichern der Ueberlappung	*/
	    for(i = 0; i < g_overlap; i++) {
		in[i] = in[i+buffsize-2*g_overlap]; 
	    }
					/* Abtransport Block o. Ueberl.	*/
	    for(i = g_overlap; i < fftsize; i++) {
		out[i] = (double)(xr[i]); 
		out[i+fftsize-g_overlap] = (double)(xi[i]); 
	    }
					/* Abtastwert speichern		*/
	    streamwrite(ostream, g_ofmt, afmt, out+g_overlap, found);
	    if (ferror(ostream)) {
		errmsg = catmsg(g_ofile, "kann nicht schreiben");
		errorhandler(errmsg);
		exit(-1);
	    }
	}
					/* Clean-Up			*/
	fclose(istream);
	fclose(cstream);
	fclose(ostream);
	fft(fftp);
	free(in);
	free(out);
	free(cfr);
	free(cfi);
	free(g_stdoutbuf);
	exit(0);
}
/*----------------------------------------------------------------------*/
/*  streamread	Feld in mit count Elementen aus stream besetzen		*/
/*		Format fmt aus stream nach double konvertieren		*/
/*		Ab EOF Feld mit Werten fill auffuellen			*/
/*		Anzahl der in stream gefundenen Werte zurueck		*/
/*----------------------------------------------------------------------*/
int streamread(stream, fmt, in, count, fill)
FILE *stream;
char *fmt;
double *in, fill;
int count;
{
	float *f;
	BYTE *b;
	WORD *w;
	int i, k;

        switch(fmt[0] | 0x20) {
	    case 'a': 
		for (i = 0; i < count; i++)
	            if (EOF == fscanf(stream, "%lf", in+i))
		        break;
	        break;
	    case 'f':
		f = (float*)in;
	       	i = FREAD(f,sizeof(float),count,stream);
		for (k = i - 1; k >= 0; k--)
		    in[k] = (double)(f[k]);
	        break;
	    case 'd':
	        i = FREAD(in,sizeof(double),count,stream);
	        break;
	    case 'b':
		b = (BYTE*)in;
	       	i = FREAD(b,sizeof(BYTE),count,stream);
		for (k = i - 1; k >= 0; k--)
		    in[k] = (double)((BYTE)(b[k]));
	        break;
	    case 'u':
		b = (BYTE*)in;
	       	i = FREAD(b,sizeof(BYTE),count,stream);
		for (k = i - 1; k >= 0; k--)
		    in[k] = (double)((unsigned BYTE)(b[k]));
	        break;
	    case 'w':
		w = (WORD*)in;
	       	i = FREAD(w,sizeof(WORD),count,stream);
		for (k = i - 1; k >= 0; k--)
		    in[k] = (double)(w[k]);
	        break;
        }
	for (k = i; k < count; k++)
		in[k] = fill;
	return(i);
}
/*----------------------------------------------------------------------*/
/*  streamwrite	Feld in mit count Elementen auf stream schreiben	*/
/*		Format double nach fmt fuer stream konvertieren		*/
/*		afmt ist der Format-String fuer die ASCII-Ausgabe	*/
/*		Anzahl der geschriebenen Werte zurueck			*/
/*----------------------------------------------------------------------*/
int streamwrite(stream, fmt, afmt, out, count)
FILE *stream;
char *fmt, *afmt;
double *out;
int count;
{
	register WORD *w;
	register int i, k;
	float *f;
	BYTE *b;
	int chunk, n, l;

	i = 0;
	switch(fmt[0] | 0x20) {
	    case 'a': 
		for (i = 0; i < count && !ferror(stream); i++)
		    fprintf(stream, afmt, out[i]);
	        break;
	    case 'f':
		f = (float*)out;
		chunk = CHUNKSIZEBYTES/sizeof(float);
		for (l = 0; l < count; l += chunk) {
		    n = (count - l < chunk) ? count - l : chunk;
		    for (k = 0; k < n; k++) 
		        f[k+l] = (float)(out[k+l]);
	       	    i += FWRITE(f+l,sizeof(float),n,stream);
		}
	        break;
	    case 'd':
		chunk = CHUNKSIZEBYTES/sizeof(double);
		for (l = 0; l < count; l += chunk) {
		    n = (count - l < chunk) ? count - l : chunk;
	       	    i += FWRITE(out+l,sizeof(double),n,stream);
		}
	        break;
	    case 'u':
		b = (BYTE*)out;
		chunk = CHUNKSIZEBYTES/sizeof(BYTE);
		for (l = 0; l < count; l += chunk) {
		    n = (count - l < chunk) ? count - l : chunk;
		    for (k = 0; k < n; k++) 
		        b[k+l] = (unsigned BYTE)(out[k+l]);
	       	    i += FWRITE(b+l,sizeof(BYTE),n,stream);
		}
		break;
	    case 'b':
		b = (BYTE*)out;
		chunk = CHUNKSIZEBYTES/sizeof(BYTE);
		for (l = 0; l < count; l += chunk) {
		    n = (count - l < chunk) ? count - l : chunk;
		    for (k = 0; k < n; k++) 
		        b[k+l] = (BYTE)(out[k+l]);
	       	    i += FWRITE(b+l,sizeof(BYTE),n,stream);
		}
	        break;
	    case 'w':
		w = (WORD*)out;
		chunk = CHUNKSIZEBYTES/sizeof(WORD);
		for (l = 0; l < count; l += chunk) {
		    n = (count - l < chunk) ? count - l : chunk;
		    for (k = 0; k < n; k++) 
		        w[k+l] = (WORD)(out[k+l]);
	       	    i += FWRITE(w+l,sizeof(WORD),n,stream);
		}
	        break;
	}
	return(i);
}
/*----------------------------------------------------------------------*/
/*  header	Hilfstext-Ueberschrift, wird von key.c gerufen.		*/
/*----------------------------------------------------------------------*/
header()
{
char string[90], version[90];

sscanf("$Revision: 2.10 $ $Date: 1999/08/04 14:00:35 $ $Author: mummert $",
        "$%*s %s $ $%*s %s %*s $ $%*s %s $",
        string, string+30, string+60);
sprintf(version, "FASTCONV V%s %s %s", string, string+30, string+60);
fprintf(stderr, "\n");
fprintf(stderr, "    %s\n\n", version);
fprintf(stderr, "    Schnelle Faltung mit FFT nach Overlap-Add-Methode");
fprintf(stderr, "\n");
fprintf(stderr, "\n");
}
/*******Ende*************************************************************/

